# -*- coding: utf-8 -*-

from sqlalchemy.orm.session import Session as BaseSession

from . import registry
from .exceptions import InvalidBindKey


class Session(BaseSession):

    def get_bind(self, mapper=None, clause=None):
        if mapper is not None:
            info = getattr(mapper.mapped_table, 'info', {})
            engine_key = info.get('bind_key')

            if engine_key is not None:
                engine = registry.get_engine(engine_key)

                if not engine:
                    raise InvalidBindKey(engine_key)

                return engine

        return BaseSession.get_bind(self, mapper, clause)



# -*- coding: utf-8 -*-

import time
import inspect
from random import randint
from functools import wraps

from ...utils import pyv
from .exceptions import ReRaiseException


def re_raise_exc(callback=None, exc_cls=ReRaiseException, message=None):
    """
    Decorator for except any exception and reraise it.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as e:
                orig_message = pyv.get_exc_message(e)

                if message:
                    new_message = u'{}{}'.format(
                        message, u' (from {}: {})'.format(e.__class__.__name__, orig_message)
                        if orig_message else
                        '(from {})'.format(e.__class__.__name__)
                    )
                else:
                    new_message = orig_message

                raise exc_cls(new_message)

        return wrapped

    if callable(callback):
        return wrapper(callback)

    return wrapper


def random_file_name(file_ex=None):
    """
    Generate random file name
    """
    file_ex = file_ex or ''
    file_name = str(
        int(time.time() + randint(0, 1000)),
    )
    file_name += file_ex
    return file_name


def change_name_from_python_to_html(name):
    """
    Attribute name from DOM tree has different
    format than python style.
    This function help with that problem.
    """
    name = name.replace('_', '-')
    if name.startswith('-'):
        return name[1:]
    if name.endswith('-'):
        return name[:-1]
    return name


def is_ready_state_complete(browser):
    """
    Do return True if document ready
    state is complete else False
    """
    state = browser.execute_script(
        'return document.readyState',
    )
    return state == 'complete'


def declare_standard_callback(func):
    """
    Check for signature of function which should
    has only one argument in signature and return it.
    """
    if not callable(func):
        return func

    try:
        signature = inspect.getargspec(func)
    except TypeError:  # if function isn't python function we can not know about it
        return func

    if 0 < len(signature.args) < 2:
        return func

    raise TypeError(
        'Incorrect signature of function "{0}" -> "{1}". Should be "{0}" -> "[\'instance\']".'.format(
            pyv.get_func_name(func), str(signature.args),

        ),
    )



# -*- coding: utf-8 -*-

from ...utils import pyv
from .exceptions import ConfigurationError


class Signature(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


def create_signatures(option_value):
    def create_signature(v):
        if isinstance(v, Signature):
            return v

        if isinstance(v, dict):
            return Signature(**v)
        elif v is True:
            return Signature()

        raise ConfigurationError(
            'Can not create signature from option value {}'.format(v),
        )

    if type(option_value) is int:
        return map(lambda i: Signature(), pyv.xrange(option_value))

    if isinstance(option_value, (list, tuple)):
        return map(create_signature, option_value)

    return [create_signature(option_value)]



# -*- coding: utf-8 -*-

import types
import inspect
from jsonschema import Draft4Validator
from jsonschema import ValidationError as JsonSchemaValidationError

from ...utils import pyv
from .exceptions import ValidationError
from .factory import FactoryMeta as FactoryType


SCHEMA = {
    'type': 'object',
    'properties': {
        'storage': {'type': 'string'},
        'embedded': {'type': 'object'},
        'initializer': {'type': 'function'},
        'factory_class': {'type': 'factory'},
        'staging': {
            'type': 'object',
            'properties': {
                'creator': {'type': 'function'},
                'pre': {'type': ['list', 'tuple']},
                'post': {'type': ['list', 'tuple']},
            },
        },
        'require': {'type': 'array'},
    },
    'required': ['factory_class'],
}


class SchemaValidator(Draft4Validator):

    DEFAULT_TYPES = Draft4Validator.DEFAULT_TYPES

    DEFAULT_TYPES.update(**{
        'list': list,
        'tuple': tuple,
        'factory': FactoryType,
        'function': types.FunctionType,
    })

    def __init__(self, schema):
        super(SchemaValidator, self).__init__(SCHEMA)
        try:
            self.validate(schema)
        except JsonSchemaValidationError as error:
            raise ValidationError(pyv.get_exc_message(error))


class SettingsValue(str):
    pass


def _get_value(schema, settings, key):
    value = schema.get(key)
    if isinstance(value, SettingsValue):
        return getattr(settings, value)
    return value


def declare_initializer_callback(initializer):
    if not callable(initializer):
        return initializer

    sig = inspect.getargspec(initializer)

    if len(sig.args) != 3:
        raise ValidationError(
            'Incorrect signature on initializer callback {} Should be {}'.format(
                sig.args, ['builder', 'cls', 'sig'],
            ),
        )

    return initializer


class RuleObject(object):

    def __init__(self, name, builder, schema, instance):
        SchemaValidator(schema)

        self._name = name
        self._schema = schema
        self._builder = builder
        self._instance = instance

        self._require = schema.get('require', [])
        self._staging = schema.get('staging', {})
        self._storage_name = schema.get('storage')
        self._property_name = schema.get('property')
        self._embedded_schema = schema.get('embedded', {})
        self._initializer = declare_initializer_callback(
            _get_value(schema, builder.settings, 'initializer'),
        )
        self._factory_class = _get_value(schema, builder.settings, 'factory_class')

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self._name)

    @property
    def name(self):
        return self._name

    @property
    def schema(self):
        return self._schema

    @property
    def builder(self):
        return self._builder

    @property
    def instance(self):
        return self._instance

    @property
    def require(self):
        return self._require

    @property
    def storage_name(self):
        return self._storage_name

    @property
    def staging(self):
        return self._staging

    @property
    def property_name(self):
        return self._property_name

    @property
    def embedded_schema(self):
        return self._embedded_schema

    @property
    def initializer(self):
        return self._initializer

    @property
    def factory_class(self):
        return self._factory_class

    @property
    def can_be_compiled(self):
        return all(r in self._instance for r in self._require)

    def get_current_requires(self):
        return [r for r in self._require if r not in self._instance]

    def compile(self, sig):
        if self._embedded_schema:
            embedded_options = dict(
                (k, sig.kwargs.pop(k))
                for k in self._embedded_schema.keys()
                if sig.kwargs.get(k) is not None
            )
        else:
            embedded_options = {}

        return CompiledRule(
            sig=sig,
            name=self._name,
            builder=self._builder,
            require=self._require,
            staging=self._staging,
            instance=self._instance,
            initializer=self._initializer,
            storage_name=self._storage_name,
            property_name=self._property_name,
            factory_class=self._factory_class,
            embedded_options=embedded_options,
            embedded_schema=self._embedded_schema,
        )


class CompiledRule(object):

    def __init__(self,
                 sig=None,
                 name=None,
                 builder=None,
                 require=None,
                 instance=None,
                 staging=None,
                 initializer=None,
                 storage_name=None,
                 property_name=None,
                 factory_class=None,
                 embedded_schema=None,
                 embedded_options=None):
        self._sig = sig
        self._name = name
        self._builder = builder
        self._require = require
        self._staging = staging
        self._instance = instance
        self._initializer = initializer
        self._storage_name = storage_name
        self._property_name = property_name
        self._factory_class = factory_class
        self._embedded_schema = embedded_schema
        self._embedded_options = embedded_options

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self._name)

    @property
    def sig(self):
        return self._sig

    @property
    def name(self):
        return self._name

    @property
    def builder(self):
        return self._builder

    @property
    def instance(self):
        return self._instance

    @property
    def require(self):
        return self._require

    @property
    def storage_name(self):
        return self._storage_name

    @property
    def staging(self):
        return self._staging

    @property
    def property_name(self):
        return self._property_name

    @property
    def embedded_schema(self):
        return self._embedded_schema

    @property
    def initializer(self):
        return self._initializer

    @property
    def factory_class(self):
        return self._factory_class

    @property
    def embedded_options(self):
        return self._embedded_options



# -*- coding: utf-8 -*-

from .proxy import WebDriverProxy
from .proxy import WebElementList
from .proxy import WebElementProxy


__all__ = (
    'WebDriverProxy',
    'WebElementList',
    'WebElementProxy',
)



# -*- coding: utf-8 -*-

import logging

from selenium.common.exceptions import WebDriverException
from selenium.common.exceptions import NoSuchElementException

from ...utils.common import waiting_for
from .exceptions import PollingTimeoutExceeded
from .utils import change_name_from_python_to_html


logger = logging.getLogger(__name__)


TAG_ALIASES = {
    'any': '*',
}

ATTRIBUTE_ALIASES = {
    '_id': 'id',
    '_class': 'class',
    '_type': 'type',
    'id_': 'id',
    'class_': 'class',
    'type_': 'type',
}


class QueryObject(object):

    def __init__(self, tag, **selector):
        self.tag = tag
        self.selector = selector

    def __call__(self, proxy):
        return make_result(proxy, self.tag)(**self.selector)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return u'<{} {}>'.format(
            self.tag,
            u' '.join(
                (u'{}="{}"'.format(k, v) for k, v in self.selector.items())
            ),
        )


class ValueFormat(object):

    __format__ = None

    def __init__(self, value):
        self.value = value

    def __call__(self, tag_attr_name=None):
        if not self.__format__:
            raise ValueError('Unknown format')

        return self.__format__.format(
            attr_value=self.value,
            attr_name=attribute_by_alias(tag_attr_name or ''),
        )


class Equal(ValueFormat):

    __format__ = u'[{attr_name}="{attr_value}"]'


class NotEqual(ValueFormat):

    __format__ = u':not([{attr_name}="{attr_value}"])'


class Contains(ValueFormat):

    __format__ = u'[{attr_name}*="{attr_value}"]'


class NotContains(ValueFormat):

    __format__ = u':not([{attr_name}*="{attr_value}"])'


class ContainsText(ValueFormat):

    __format__ = u':contains("{attr_value}")'


class StartsWith(ValueFormat):

    __format__ = u'[{attr_name}^="{attr_value}"]'


class EndsWith(ValueFormat):

    __format__ = u'[{attr_name}$="{attr_value}"]'


class Expression(object):

    __restricted_formats__ = (
        ContainsText,
    )

    def __init__(self, value=None, value_format=Equal):
        self._expressions = []

        if value:
            self.expression(value, value_format=value_format)

    def expression(self, value, value_format=Equal):
        if isinstance(value, ValueFormat):
            self._add_format(value)
        else:
            self._add_format(value_format(value))

        return self

    def _check_format(self, fmt):
        if fmt.__class__ in self.__restricted_formats__:
            raise ValueError(
                'Restricted format "{}"'.format(fmt.__class__.__name__),
            )

    def _add_format(self, fmt):
        self._check_format(fmt)
        self._expressions.append(fmt)

    def __call__(self, tag_attr_name):
        css_selector = []

        for fmt in self._expressions:
            css_selector.append(fmt(tag_attr_name))

        return u''.join(css_selector)


def tag_by_alias(tag_name):
    return TAG_ALIASES.get(tag_name, tag_name)


def attribute_by_alias(atr_name):
    return change_name_from_python_to_html(
        ATTRIBUTE_ALIASES.get(atr_name, atr_name),
    )


def make_result(proxy, tag):
    def handle(tag_text_contains=None, **selector):
        css = [tag_by_alias(tag)]

        if isinstance(tag_text_contains, ContainsText):
            css.append(tag_text_contains())

        for tag_attr_name, value in selector.items():
            if isinstance(value, Expression):
                expression = value
            else:
                expression = Expression(value)

            css.append(expression(tag_attr_name))

        return QueryResult(proxy, u''.join(css))

    return handle


def get_execute_method(proxy, list_result):
    css_executors = {
        True: 'find_elements_by_css_selector',
        False: 'find_element_by_css_selector',
    }
    method_name = css_executors[bool(list_result)]

    return getattr(proxy, method_name)


def execute(proxy, css, list_result=False, disable_polling=False):
    logger.debug(
        u'Execute css query "{}", list result "{}"'.format(
            css, 'Yes' if list_result else 'No',
        ),
    )

    proxy.reason_storage['last css query'] = css

    if disable_polling:
        with proxy.disable_polling():
            wait_timeout = proxy.config.WAIT_TIMEOUT
            proxy.config.WAIT_TIMEOUT = 0
            try:
                method = get_execute_method(proxy, list_result)
                result = method(css)
            finally:
                proxy.config.WAIT_TIMEOUT = wait_timeout
        return result

    method = get_execute_method(proxy, list_result)
    return method(css)


class QueryResult(object):

    def __init__(self, proxy, css):
        self.__we = None

        self.__css = css
        self.__proxy = proxy

    def __repr__(self):
        return u'{}({}): {}'.format(
            self.__class__.__name__, self.__css, repr(self.__proxy),
        )

    def __getattr__(self, item):
        if not self.__we:
           self.first()
        return getattr(self.__we, item)

    def __css_string__(self):
        return self.__css

    @property
    def exist(self):
        """
        Check exist first element of query

        :rtype: bool
        """
        try:
            el = execute(self.__proxy, self.__css, disable_polling=True)

            if el:
                return True
            return False
        except WebDriverException:
            return False

    def wait(self, timeout=None, delay=None):
        """
        Wait for first element of query while timeout doesn't exceeded

        :param timeout: time for wait in seconds
        """
        return waiting_for(
            lambda: self.exist,
            exc_cls=PollingTimeoutExceeded,
            delay=delay or self.__proxy.config.POLLING_DELAY,
            timeout=timeout or self.__proxy.config.POLLING_TIMEOUT,
            message='Could not wait web element by css "{}"'.format(self.__css),
        )

    def get(self, index):
        """
        Get element of query by index

        :param index: index of element
        """
        try:
            self.__we = execute(self.__proxy, self.__css, list_result=True)[index]
            return self.__we
        except IndexError:
            raise NoSuchElementException(
                'Result does not have element by index "{}". Css query: "{}".'.format(
                    index, self.__css,
                ),
            )

    def first(self):
        """
        Get first element of query
        """
        self.__we = execute(self.__proxy, self.__css)
        return self.__we

    def all(self):
        """
        Get all elements of query
        """
        return execute(self.__proxy, self.__css, list_result=True)


class Query(object):
    """
    Make query data for execution.

    q = query(query.DIV, _class=query.contains('name'))
    result = q(browser)
    result.first()
    """

    A = 'a'
    B = 'b'
    P = 'p'
    U = 'u'
    UL = 'ul'
    LI = 'li'
    BR = 'br'
    EM = 'em'
    HR = 'hr'
    TR = 'tr'
    TD = 'td'
    TH = 'th'
    TT = 'tt'
    VAR = 'var'
    IMG = 'img'
    DIV = 'div'
    MAP = 'map'
    HEAD = 'head'
    FORM = 'form'
    BODY = 'body'
    AREA = 'area'
    CODE = 'code'
    BASE = 'base'
    SPAN = 'span'
    LINK = 'link'
    META = 'meta'
    SMALL = 'small'
    TABLE = 'table'
    INPUT = 'input'
    LABEL = 'label'
    FRAME = 'frame'
    EMBED = 'embed'
    BLINK = 'blink'
    IFRAME = 'iframe'
    CENTER = 'center'
    STRONG = 'strong'
    BUTTON = 'button'
    OBJECT = 'object'
    OPTION = 'option'
    SELECT = 'select'
    TEXTAREA = 'textarea'
    OPTGROUP = 'optgroup'
    FRAMESET = 'frameset'

    ANY = '*'

    equal = Equal
    not_equal = NotEqual
    contains = Contains
    endswith = EndsWith
    startswith = StartsWith
    not_contains = NotContains
    contains_text = ContainsText

    @staticmethod
    def css_string(query_result):
        return query_result.__css_string__()

    @staticmethod
    def expression(*args):
        expression = Expression()

        for fmt in args:
            expression.expression(fmt)

        return expression

    def __call__(self, tag, **selector):
        return QueryObject(tag, **selector)


query = Query()



# -*- coding: utf-8 -*-

from selenium.webdriver.common.alert import Alert as _Alert
from selenium.webdriver.common.action_chains import ActionChains as _ActionChains
from selenium.webdriver.common.touch_actions import TouchActions as _TouchActions


class Alert(_Alert):

    def __init__(self, proxy):
        super(Alert, self).__init__(proxy.browser)

    @property
    def browser(self):
        return self.driver


class ActionChains(_ActionChains):

    def __init__(self, proxy):
        super(ActionChains, self).__init__(proxy.browser)

    @property
    def browser(self):
        return self._driver

    def reset(self):
        self._actions = []

    def __exit__(self, exc_type, exc_val, exc_tb):
        super(ActionChains, self).__exit__(exc_type, exc_val, exc_tb)
        self.reset()


class TouchActions(_TouchActions):

    def __init__(self, proxy):
        super(TouchActions, self).__init__(proxy.browser)

    @property
    def browser(self):
        return self._driver

    def reset(self):
        self._actions = []

    def __exit__(self, exc_type, exc_val, exc_tb):
        super(TouchActions, self).__exit__(exc_type, exc_val, exc_tb)
        self.reset()



# -*- coding: utf-8 -*-

from . import signature
from . import exceptions
from .rule import RuleObject
from .staging import Staging


class RelationObject(object):

    def __init__(self, *keys):
        self._keys = []
        self._keys.extend(keys)

    def __getitem__(self, item):
        self._keys.append(item)
        return self

    def __repr__(self):
        return 'Relation to schema: {}'.format('->'.join(self._keys))

    def __call__(self, builder):
        try:
            data = builder.__build_schema__[self._keys[0]]
        except KeyError:
            raise exceptions.RelationError(
                '{} from {}'.format(self._keys[0], '->'.join(self._keys))
            )

        for key in self._keys[1:]:
            try:
                data = data['embedded'][key]
            except KeyError:
                raise exceptions.RelationError(
                    '{} from {}'.format(key, '->'.join(self._keys)),
                )

        return data


class CurrentSchema(object):

    def __getitem__(self, item):
        return RelationObject(item)


current_schema = CurrentSchema()


class PendingList(object):

    def __init__(self):
        self.__lst = []

    def add(self, rule_object, option_value):
        self.__lst.append((rule_object, option_value))

    def delete(self, rule_object, option_value):
        self.__lst.remove((rule_object, option_value))

    def items(self):
        return tuple(self.__lst)

    def unlocked_items(self):
        return tuple(filter(lambda i: i[0].can_be_compiled, self.__lst))

    def locked_items(self):
        return tuple(filter(lambda i: not i[0].can_be_compiled, self.__lst))

    def join(self):
        iteration = 0
        length = len(self.__lst)

        while self.__lst:
            for rule_object, option_value in self.unlocked_items():
                for sig in signature.create_signatures(option_value):
                    create_schema(rule_object.compile(sig))

                self.delete(rule_object, option_value)

            iteration += 1

            if iteration > length:
                errors = []

                for rule_object, _ in self.locked_items():
                    errors.append(
                        'Builder schema "{}" can not be collected. Was required {} but does not created.'.format(
                            rule_object.name, ', '.join('"{}"'.format(r) for r in rule_object.get_current_requires())
                        ),
                    )

                    raise exceptions.SchemaError('\n'.join(errors))


def create_schema(compiled_rule):
    staging = Staging(compiled_rule)

    staging.pre()
    staging.create()
    staging.post()

    if compiled_rule.embedded_options:
        collect(
            compiled_rule.builder,
            schema=compiled_rule.embedded_schema,
            options=compiled_rule.embedded_options,
            instance=staging.embedded_instance,
        )


def collect(builder, schema=None, options=None, instance=None):
    pending_list = PendingList()

    instance = instance or builder.schema
    schema = schema or builder.__build_schema__
    options = options or builder.settings.options

    for option_name, option_value in options.items():
        try:
            option_schema = schema[option_name]
        except KeyError:
            raise exceptions.ConfigurationError(
                'Incorrect option "{}". Builder has not schema for it.'.format(option_name),
            )

        if isinstance(option_schema, RelationObject):
            option_schema = option_schema(builder)

        rule_object = RuleObject(option_name, builder, option_schema, instance)

        if rule_object.can_be_compiled:
            for sig in signature.create_signatures(option_value):
                create_schema(rule_object.compile(sig))
        else:
            pending_list.add(rule_object, option_value)

    pending_list.join()



# -*- coding: utf-8 -*-

import sys
from functools import wraps

from . import loader
from . import runnable
from .utils import pyv


STEP_ATTRIBUTE_NAME = '__step__'
STEP_DOC_ATTRIBUTE_NAME = '__doc__'
STEP_WEIGHT_ATTRIBUTE_NAME = '__weight__'
STEPS_HISTORY_ATTRIBUTE_NAME = '__history__'
CURRENT_STEP_ATTRIBUTE_NAME = '__current_step__'
CURRENT_FLOW_ATTRIBUTE_NAME = '__current_flow__'
STEP_BY_STEP_ATTRIBUTE_NAME = '__step_by_step__'
STEPS_STORAGE_ATTRIBUTE_NAME = '__step_methods__'


def step(num, doc=None, performer=None):
    def wrapper(method):
        if type(num) not in (int, float):
            raise ValueError('step num can be int or float type only')

        setattr(method, STEP_DOC_ATTRIBUTE_NAME, doc or '')
        setattr(method, STEP_WEIGHT_ATTRIBUTE_NAME, num)
        setattr(method, STEP_ATTRIBUTE_NAME, True)

        @wraps(method)
        def wrapped(self, *args, **kwargs):
            if performer:
                step_method = lambda: method(self, *args, **kwargs)
                return performer(self, step_method)
            return method(self, *args, **kwargs)

        return wrapped

    return wrapper


def get_step_methods(case):
    return getattr(case, STEPS_STORAGE_ATTRIBUTE_NAME)


def get_case_name(case):
    return case.__class__.__name__


def get_step_doc(method):
    return pyv.unicode_string(
        getattr(method, STEP_DOC_ATTRIBUTE_NAME),
    )


def is_step_by_step_case(case):
    return hasattr(case, STEP_BY_STEP_ATTRIBUTE_NAME)


def get_current_step(case):
    return pyv.unicode_string(
        getattr(case, CURRENT_STEP_ATTRIBUTE_NAME),
    )


def get_current_flow(case):
    return getattr(case, CURRENT_FLOW_ATTRIBUTE_NAME)


def get_step_num(method):
    return getattr(method, STEP_WEIGHT_ATTRIBUTE_NAME)


def get_case_history(case):
    return getattr(case, STEPS_HISTORY_ATTRIBUTE_NAME)


def _create_history_line(method):
    return u'{}. {}'.format(
        get_step_num(method),
        get_step_doc(method),
    )


def _create_current_step_info(case, method):
    return u'<{}.{}({}): {}>'.format(
        get_case_name(case),
        pyv.get_func_name(method),
        get_step_num(method),
        get_step_doc(method),
    )


def _step_log(case, method, flow):
    doc = get_step_doc(method)
    num = get_step_num(method)
    case_name = get_case_name(case)
    method_name = pyv.get_func_name(method)

    return u'{}{}. {}.{}: "{}"'.format(
        ''.join(' ' for _ in pyv.xrange(4 if flow else 2)),
        num, case_name, method_name, doc,
    )


def _perform_prompt(case, method, exit_code=None):
    commands = {  # prompt commands
        'exit': 'q',
        'continue': 'c',
    }
    prompt = '>> {}.{} [{}]: '.format(
        case.__class__.__name__,
        pyv.get_func_name(method),
        ', '.join([c for _, c in commands.items()]),
    )
    command = case.log.prompt(prompt).strip()

    if command == commands['exit']:
        sys.exit(exit_code or 0)


def _run_step(case, method, flow=None):
    if case.config.STEP_BY_STEP:
        _perform_prompt(case, method)

    if case.config.STEPS_LOG:
        case.log(_step_log(case, method, flow))

    try:
        if flow is not None:
            method(case, flow)
        else:
            method(case)
    except BaseException:
        runnable.stopped_on(case, pyv.get_func_name(method))
        raise


def _call_to_begin_method_if_exist(case, flow=None):
    if hasattr(case, 'begin'):
        setattr(
            case,
            CURRENT_STEP_ATTRIBUTE_NAME,
            'begin',
        )
        if flow:
            case.begin(flow)
        else:
            case.begin()


def _call_to_finish_method_if_exist(case, flow=None):
    if hasattr(case, 'finish'):
        setattr(
            case,
            CURRENT_STEP_ATTRIBUTE_NAME,
            'finish',
        )
        if flow:
            case.finish(flow)
        else:
            case.finish()


def _make_run_test():
    def run_test(self):
        run_test.__doc__ = self.__doc__

        if self.__flows__:

            for flow in self.__flows__:
                _call_to_begin_method_if_exist(self, flow=flow)

                setattr(self, CURRENT_FLOW_ATTRIBUTE_NAME, flow)

                if self.config.STEPS_LOG:
                    self.log(u'  Flow: ', pyv.unicode_string(flow))

                for step_method in get_step_methods(self):
                    setattr(
                        self,
                        CURRENT_STEP_ATTRIBUTE_NAME,
                        _create_current_step_info(self, step_method),
                    )
                    getattr(self, STEPS_HISTORY_ATTRIBUTE_NAME).append(
                        _create_history_line(step_method),
                    )

                    _run_step(self, step_method, flow=flow)
                else:
                    setattr(self, STEPS_HISTORY_ATTRIBUTE_NAME, [])

                _call_to_finish_method_if_exist(self, flow=flow)

        else:
            _call_to_begin_method_if_exist(self)

            for step_method in get_step_methods(self):
                setattr(
                    self,
                    CURRENT_STEP_ATTRIBUTE_NAME,
                    _create_current_step_info(self, step_method),
                )
                getattr(self, STEPS_HISTORY_ATTRIBUTE_NAME).append(
                    _create_history_line(step_method),
                )

                _run_step(self, step_method)

            _call_to_finish_method_if_exist(self)

    return run_test


class CaseMeta(type):

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)

        steps = []

        attributes = (
            a for a in dir(cls)
            if not a.startswith('_')
        )

        for atr in attributes:
            method = getattr(cls, atr, None)
            if hasattr(method, STEP_ATTRIBUTE_NAME):
                steps.append(method)

        if steps:
            steps.sort(
                key=lambda m: getattr(m, STEP_WEIGHT_ATTRIBUTE_NAME),
            )

            setattr(cls, STEPS_HISTORY_ATTRIBUTE_NAME, [])
            setattr(cls, CURRENT_STEP_ATTRIBUTE_NAME, None)
            setattr(cls, STEP_BY_STEP_ATTRIBUTE_NAME, True)
            setattr(cls, STEPS_STORAGE_ATTRIBUTE_NAME, steps)
            setattr(cls, CURRENT_FLOW_ATTRIBUTE_NAME, 'Without flows')
            setattr(cls, loader.DEFAULT_TEST_NAME, _make_run_test())

        return cls



# -*- coding: utf-8 -*-

import sys
from contextlib import contextmanager


@contextmanager
def dev_null():
    class MockStd(object):

        def __getattr__(self, item):
            def null(*args, **kwargs):
                pass
            return null

    stdout = sys.stdout
    stderr = sys.stderr
    sys.stdout = MockStd()
    sys.stderr = MockStd()
    try:
        yield
    finally:
        sys.stdout = stdout
        sys.stderr = stderr



# -*- coding: utf-8 -*-

import time
from functools import wraps
from types import MethodType
from collections import OrderedDict
from contextlib import contextmanager

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement

from . import tools
from . import actions
from .. import polling
from ....utils import pyv
from ..router import Router
from ..query import make_result
from ....utils.common import waiting_for
from .interfaces import WebDriverInterface
from .interfaces import WebElementInterface
from ..utils import is_ready_state_complete
from .interfaces import HTML_TAGS_ALLOWED_AS_METHOD


def factory_method(f, browser, config, allow_polling=True, reason_storage=None):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)

        if isinstance(result, WebElement):
            return WebElementProxy(
                result,
                config=config,
                browser=browser,
                allow_polling=allow_polling,
                reason_storage=reason_storage,
            )

        if isinstance(result, (list, tuple)):
            we_list = WebElementList(
                browser,
                config,
                reason_storage=reason_storage,
            )

            for obj in result:
                if isinstance(obj, WebElement):
                    we_list.append(
                        WebElementProxy(
                            obj,
                            browser=browser,
                            config=config,
                            allow_polling=allow_polling,
                            reason_storage=reason_storage,
                        ),
                    )
                else:
                    return result

            return we_list
        return result
    return wrapper


class WebElementList(list):

    def __init__(self, browser, config, reason_storage=None):
        super(WebElementList, self).__init__()

        self.__browser = browser
        self.__config = config
        self.__reason_storage = reason_storage or OrderedDict()

    @property
    def browser(self):
        return self.__browser

    @property
    def config(self):
        return self.__config

    @property
    def reason_storage(self):
        return self.__reason_storage

    def get_by(self, **kwargs):
        try:
            return next(self.filter(**kwargs))
        except StopIteration:
            pass

    def filter(self, **kwargs):
        def check_equal(we, attr, value):
            attr_value = getattr(we, attr, None)
            if callable(attr_value) or attr_value is None:
                attr_value = getattr(we.attr, attr)
            return attr_value == value

        for we in self:
            if all(check_equal(we, k, v) for k, v in kwargs.items()):
                yield we


class BaseProxy(object):

    def __init__(self, wrapped, config=None, browser=None, reason_storage=None, allow_polling=True):
        self.__browser = browser
        self.__config = config
        self.__wrapped = wrapped
        self._allow_polling = allow_polling
        self.__reason_storage = reason_storage or OrderedDict()

        assert self.__class__ != BaseProxy, 'This is base class only. Can not be instanced.'

    def __len__(self):
        if isinstance(self._wrapped, WebElement):
            area = self
        else:
            area = self.body().first()

        try:
            return len(area.attr.innerHTML)
        except (TypeError, WebDriverException):
            return 0

    def __nonzero__(self):
        return True

    def __bool__(self):  # please python 3
        return self.__nonzero__()

    def __dir__(self):
        return list((dir(self._wrapped) + dir(self.__class__)))

    def __repr__(self):
        return repr(self._wrapped)

    def __getattr__(self, item):
        if item in HTML_TAGS_ALLOWED_AS_METHOD:
            return make_result(self, item)

        raise AttributeError(
            '"{}" has not attribute "{}"'.format(
                self.__class__.__name__, item,
            ),
        )

    def __getattr_from_webdriver_or_webelement__(self, item):
        attr = getattr(self._wrapped, item)

        if callable(attr) and type(attr) == MethodType:
            if self.allow_polling:
                return polling.do(
                    factory_method(
                        attr,
                        self.browser,
                        self.config,
                        allow_polling=self.allow_polling,
                        reason_storage=self.reason_storage,
                    ),
                    delay=self.config.POLLING_DELAY,
                    timeout=self.config.POLLING_TIMEOUT,
                )

            return factory_method(
                attr,
                self.browser,
                self.config,
                allow_polling=self.allow_polling,
                reason_storage=self.reason_storage,
            )

        return attr

    def __setattr_to_webdriver_or_webelement__(self, item, value):
        setattr(self._wrapped, item, value)

    @property
    def _wrapped(self):
        return self.__wrapped

    @property
    def config(self):
        return self.__config

    @property
    def browser(self):
        return self.__browser

    @property
    def allow_polling(self):
        return (
            bool(self.config.POLLING_TIMEOUT) and self._allow_polling
        )

    @property
    def reason_storage(self):
        return self.__reason_storage

    @property
    def text(self):
        def get_text():
            if isinstance(self._wrapped, WebElement):
                return self._wrapped.text
            return self._wrapped.find_element_by_tag_name('body').text

        if self.allow_polling:
            func = polling.do(
                get_text,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return get_text()

    @property
    def is_web_element(self):
        raise NotImplementedError(
            'Property "is_web_element" does not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )

    @contextmanager
    def do_polling(self,
                   callback,
                   exceptions=None,
                   timeout=None,
                   delay=None):
        return polling.do(
            callback,
            exceptions=exceptions or polling.POLLING_EXCEPTIONS,
            delay=delay or self.config.POLLING_DELAY,
            timeout=timeout or self.config.POLLING_TIMEOUT,
        )

    def waiting_for(self,
                    callback,
                    timeout=None,
                    exc_cls=None,
                    message=None,
                    delay=None,
                    args=None,
                    kwargs=None):
        delay = delay or self.config.POLLING_DELAY
        timeout = timeout or self.config.POLLING_TIMEOUT
        exc_cls = exc_cls or polling.PollingTimeoutExceeded
        message = message or 'Wait timeout "{}" has been exceeded'.format(timeout)

        return waiting_for(
            callback,
            args=args,
            kwargs=kwargs,
            delay=delay,
            timeout=timeout,
            exc_cls=exc_cls,
            message=message,
        )

    @contextmanager
    def disable_polling(self):
        try:
            self._allow_polling = False
            yield
        finally:
            self._allow_polling = True

    def wait_ready(self, tries=15, delay=0.01):
        waiting_for(
            is_ready_state_complete,
            args=(self.browser, ),
            delay=self.config.POLLING_DELAY,
            timeout=self.config.POLLING_TIMEOUT,
            message='Timeout waiting for load page',
        )

        def is_update():
            length = [
                (len(self), time.sleep(delay), len(self))
                for _ in pyv.xrange(tries)
            ]
            statuses = [d[0] < d[2] for d in length]
            return True in statuses or length[0][0] != length[-1:][0][2]

        while is_update():
            time.sleep(delay)


class WebElementProxy(BaseProxy, WebElementInterface):

    def __init__(self, *args, **kwargs):
        super(WebElementProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebElement), 'This is proxy to WebElement only'

    @property
    def css(self):
        return tools.WebElementCssToObject(self)

    @property
    def attr(self):
        return tools.WebElementToObject(self, allow_raise=False)

    @property
    def is_web_element(self):
        return True

    def double_click(self):
        with self.browser.action_chains as action:
            action.double_click(self)
            action.perform()

    def context_click(self):
        with self.browser.action_chains as action:
            action.context_click(self)
            action.perform()


class WebDriverProxy(BaseProxy, WebDriverInterface):

    def __init__(self, *args, **kwargs):
        super(WebDriverProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebDriver), 'This is proxy to WebDriver only'

        self.__router = Router(self)
        self.__alert = actions.Alert(self)
        self.__action_chains = actions.ActionChains(self)
        self.__touch_actions = actions.TouchActions(self)

    @property
    def browser(self):
        return self

    @property
    def router(self):
        return self.__router

    @property
    def alert(self):
        return self.__alert

    @property
    def action_chains(self):
        return self.__action_chains

    @property
    def touch_actions(self):
        return self.__touch_actions

    @property
    def current_url(self):
        if self.allow_polling:
            func = polling.do(
                lambda: self._wrapped.current_url,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return self._wrapped.current_url

    @property
    def current_path(self):
        if not self.config.PROJECT_URL:
            raise RuntimeError(
                'Can not calculate current path. PROJECT_URL is not set.',
            )
        return self.current_url.replace(self.config.PROJECT_URL, '')

    @property
    def is_web_element(self):
        return False



# -*- coding: utf-8 -*-

from .exceptions import InvalidBindKey


_ENGINES = {}

_SESSION = None


def register_engine(bind_key, engine):
    _ENGINES[bind_key] = engine


def get_engine(bind_key):
    try:
        return _ENGINES[bind_key]
    except KeyError:
        raise InvalidBindKey(bind_key)


def register_session(session):
    global _SESSION
    _SESSION = session


def get_session():
    if _SESSION is None:
        raise RuntimeError('working outside session scope')

    return _SESSION



# -*- coding: utf-8 -*-

import inspect

from . import signature
from ...utils import pyv
from .exceptions import BuildError
from .exceptions import ValidationError


class AliasToMethod(object):

    def __init__(self, option_name, method_name):
        self._option_name = option_name
        self._method_name = method_name

    @property
    def option_name(self):
        return self._option_name

    @property
    def method_name(self):
        return self._method_name


class FactoryMethodWrapper(object):

    def __init__(self, parent_sig, option_name, method_name):
        self._option_name = option_name
        self._method_name = method_name
        self._option_value = parent_sig.kwargs.pop(self._option_name, None)

    def __repr__(self):
        return '{}: {}->{}'.format(
            self.__class__.__name__, self._option_name, self._method_name,
        )

    def __call__(self, factory_instance):
        if self._option_value is None:
            return

        method = getattr(factory_instance, self._method_name)

        for sig in signature.create_signatures(self._option_value):
            method(*sig.args, **sig.kwargs)


def _approve_alias(data):
    if isinstance(data, AliasToMethod):
        return data
    if isinstance(data, (list, tuple)):
        option_name, method_name = data
        return AliasToMethod(option_name, method_name)
    elif isinstance(data, pyv.basestring):
        return AliasToMethod(data, data)

    raise BuildError('Incorrect pre/post staging data "{}"'.format(data))


def create_factory_methods(sig, staging):
    pre_methods = []
    post_methods = []

    for alias in staging.get('pre', []):
        alias = _approve_alias(alias)
        pre_methods.append(
            FactoryMethodWrapper(sig, alias.option_name, alias.method_name),
        )

    for alias in staging.get('post', []):
        alias = _approve_alias(alias)
        post_methods.append(
            FactoryMethodWrapper(sig, alias.option_name, alias.method_name),
        )

    return pre_methods, post_methods


def call_to_methods(methods, factory_instance):
    for method_wrapper in methods:
        method_wrapper(factory_instance)


def declare_creator_callback(creator):
    if not callable(creator):
        return creator

    sig = inspect.getargspec(creator)

    if len(sig.args) != 2:
        raise ValidationError(
            'Incorrect signature on creator callback {} Should be {}'.format(
                sig.args, ['factory', 'schema'],
            ),
        )

    return creator


class Staging(object):

    def __init__(self, compiled_rule):
        self._pre_methods, self._post_methods = create_factory_methods(
            compiled_rule.sig, compiled_rule.staging,
        )

        if callable(compiled_rule.initializer):
            self._factory = compiled_rule.initializer(
                compiled_rule.builder, compiled_rule.factory_class, compiled_rule.sig,
            )
        else:
            self._factory = compiled_rule.factory_class(
                *compiled_rule.sig.args, **compiled_rule.sig.kwargs
            )

        self._embedded_instance = None
        self._compiled_rule = compiled_rule

    @property
    def factory(self):
        return self._factory

    @property
    def embedded_instance(self):
        return self._embedded_instance

    def pre(self):
        call_to_methods(self._pre_methods, self._factory)

    def create(self):
        creator = declare_creator_callback(
            self._compiled_rule.staging.get('creator'),
        )

        if callable(creator):
            originated = creator(self._factory, self._compiled_rule.instance)
        else:
            originated = self._factory.create()

        self._embedded_instance = self._compiled_rule.builder.__schema_class__(
            self._compiled_rule.name, originated, self._factory,
        )
        instance_name = self._compiled_rule.property_name or self._compiled_rule.name
        self._compiled_rule.instance[instance_name] = self._embedded_instance

        if self._compiled_rule.storage_name:
            if self._compiled_rule.storage_name not in self._compiled_rule.instance:
                self._compiled_rule.instance[self._compiled_rule.storage_name] = []

            self._compiled_rule.instance[self._compiled_rule.storage_name].append(self._embedded_instance)

    def post(self):
        call_to_methods(self._post_methods, self._factory)



# -*- coding: utf-8 -*-

from contextlib import contextmanager

from . import registry
from .constants import DEFAULT_BIND_KEY


EX_NAME = 'db'


class DBClient(object):
    """
    Read-write client
    """

    def __call__(self, bind_key=DEFAULT_BIND_KEY, model=None):
        if model:
            return model.bind_key(bind_key)
        return self.get_connection(bind_key)

    @staticmethod
    def get_connection(bind_key=DEFAULT_BIND_KEY):
        return registry.get_engine(bind_key).connect()

    @contextmanager
    def read(self, bind_key=DEFAULT_BIND_KEY):
        yield self.get_connection(bind_key).execute

    @contextmanager
    def write(self, bind_key=DEFAULT_BIND_KEY):
        connection = self.get_connection(bind_key)
        trans = connection.begin()
        try:
            yield connection.execute
            trans.commit()
        except BaseException as error:
            trans.rollback()
            raise error



# -*- coding: utf8 -*-

import logging

try:
    from urllib import urlencode
except ImportError:  # please python 3
    from urllib.parse import urlencode

from sqlalchemy import event
from sqlalchemy.pool import NullPool
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.result import exc
from sqlalchemy.orm import scoped_session

from . import registry
from .session import Session
from .constants import DEFAULT_BIND_KEY
from .constants import DEFAULT_POOL_CLASS


logger = logging.getLogger(__name__)


DEFAULT_AUTO_FLUSH = True
DEFAULT_AUTO_COMMIT = False
DEFAULT_EXPIRE_ON_COMMIT = True


def ping_connection(connection, connection_record, connection_proxy):
    cursor = connection.cursor()

    try:
        cursor.execute('SELECT 1')
    except:
        raise exc.DisconnectionError()

    cursor.close()


def setup_engine(
        protocol,
        host,
        port,
        db,
        user,
        password=None,
        dns_params=None,
        bind_key=DEFAULT_BIND_KEY,
        pool_class=DEFAULT_POOL_CLASS,
        **engine_options):
    """
    :param host: db host
    :param port: db port
    :param db: db name
    :param user: user name
    :param password: user password
    :param protocol: dns protocol
    :param pool_class: pool class (original param name: poolclass)
    :param dns_params: params to dns
    :param engine_options: engine kwargs
    """
    logger.debug('Setup engine for database "{}"'.format(db))

    if dns_params and isinstance(dns_params, dict):
        dns_params = '?' + urlencode(dns_params)
    else:
        dns_params = ''

    dns = '{protocol}://{user}:{password}@{host}:{port}/{database}' + dns_params
    dns = dns.format(
        user=user,
        host=host,
        port=port,
        database=db,
        protocol=protocol,
        password=password or '',
    )
    pool_class = pool_class or DEFAULT_POOL_CLASS

    logger.debug('DNS is "{}"'.format(dns))
    logger.debug('Bind key is "{}"'.format(bind_key))
    logger.debug('Pool class is "{}"'.format(pool_class.__name__))

    engine = create_engine(
        dns,
        poolclass=pool_class,
        **engine_options
    )

    if pool_class != NullPool:
        event.listen(engine, 'checkout', ping_connection)

    registry.register_engine(bind_key, engine)


def setup_session(
        autoflush=DEFAULT_AUTO_FLUSH,
        autocommit=DEFAULT_AUTO_COMMIT,
        expire_on_commit=DEFAULT_EXPIRE_ON_COMMIT,
        **options):
    """
    Setup orm session
    """
    logger.debug('Setup orm session')

    sess = scoped_session(
        sessionmaker(
            class_=Session,
            autoflush=autoflush,
            autocommit=autocommit,
            expire_on_commit=expire_on_commit,
            **options
        ),
    )
    registry.register_session(sess)



# -*- coding: utf-8 -*-

"""
Usage:
    python -m seismograph <suites path> [options]
"""

import sys
from importlib import import_module


def apply_gevent_patch():
    try:
        from gevent.monkey import patch_all
    except ImportError as e:
        from seismograph.utils import pyv
        pyv.check_gevent_supported()
        raise e

    patch_all()


def get_suites_path():
    try:
        suites_path = sys.argv[1]
        if not suites_path.startswith('-'):
            sys.argv.remove(suites_path)
            return suites_path
    except IndexError:
        pass
    return None


def get_user_main():
    try:
        main_path = sys.argv[1]
        if not main_path.startswith('-') and ':' in main_path:
            import_path, function_name = main_path.split(':')
            user_main = getattr(import_module(import_path), function_name)
            sys.argv.remove(main_path)
            return user_main
    except IndexError:
        pass
    return None


def main():
    user_main = get_user_main()

    if not user_main and '--gevent' in sys.argv:
        apply_gevent_patch()

    if user_main:
        user_main()
    else:
        import seismograph

        seismograph.main(suites_path=get_suites_path())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from .base import BaseInterface


class WebElementInterface(BaseInterface):

    def __hash__(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('__hash__')(*args, **kwargs)

    def __eq__(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('__eq__')(*args, **kwargs)

    def __ne__(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('__ne__')(*args, **kwargs)

    @property
    def tag_name(self):
        return self.__getattr_from_webdriver_or_webelement__('tag_name')

    @property
    def text(self):
        return self.__getattr_from_webdriver_or_webelement__('text')

    @property
    def location_once_scrolled_into_view(self):
        return self.__getattr_from_webdriver_or_webelement__('location_once_scrolled_into_view')

    @property
    def size(self):
        return self.__getattr_from_webdriver_or_webelement__('size')

    @property
    def location(self):
        return self.__getattr_from_webdriver_or_webelement__('location')

    @property
    def rect(self):
        return self.__getattr_from_webdriver_or_webelement__('rect')

    @property
    def screenshot_as_base64(self):
        return self.__getattr_from_webdriver_or_webelement__('screenshot_as_base64')

    @property
    def screenshot_as_png(self):
        return self.__getattr_from_webdriver_or_webelement__('screenshot_as_png')

    @property
    def parent(self):
        return self.__getattr_from_webdriver_or_webelement__('parent')

    @property
    def id(self):
        return self.__getattr_from_webdriver_or_webelement__('id')

    def _execute(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('_execute')(*args, **kwargs)

    def _upload(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('_upload')(*args, **kwargs)

    def click(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('click')(*args, **kwargs)

    def submit(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('submit')(*args, **kwargs)

    def clear(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('clear')(*args, **kwargs)

    def get_attribute(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_attribute')(*args, **kwargs)

    def is_selected(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('is_selected')(*args, **kwargs)

    def is_enabled(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('is_enabled')(*args, **kwargs)

    def find_element_by_id(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_id')(*args, **kwargs)

    def find_elements_by_id(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_id')(*args, **kwargs)

    def find_element_by_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_name')(*args, **kwargs)

    def find_elements_by_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_name')(*args, **kwargs)

    def find_element_by_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_link_text')(*args, **kwargs)

    def find_elements_by_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_link_text')(*args, **kwargs)

    def find_element_by_partial_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_partial_link_text')(*args, **kwargs)

    def find_elements_by_partial_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_partial_link_text')(*args, **kwargs)

    def find_element_by_tag_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_tag_name')(*args, **kwargs)

    def find_elements_by_tag_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_tag_name')(*args, **kwargs)

    def find_element_by_xpath(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_xpath')(*args, **kwargs)

    def find_elements_by_xpath(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_xpath')(*args, **kwargs)

    def find_element_by_class_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_class_name')(*args, **kwargs)

    def find_elements_by_class_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_class_name')(*args, **kwargs)

    def find_element_by_css_selector(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_css_selector')(*args, **kwargs)

    def find_elements_by_css_selector(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_css_selector')(*args, **kwargs)

    def send_keys(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('send_keys')(*args, **kwargs)

    def set(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('send_keys')(*args, **kwargs)

    def is_displayed(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('is_displayed')(*args, **kwargs)

    def value_of_css_property(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('value_of_css_property')(*args, **kwargs)

    def screenshot(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('screenshot')(*args, **kwargs)

    def find_element(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element')(*args, **kwargs)

    def find_elements(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements')(*args, **kwargs)



# -*- coding: utf-8 -*-

from .base import BaseInterface


class CommonWebDriverInterface(BaseInterface):
    """
    Common interface for all drivers
    """

    @property
    def mobile(self):
        return self.__getattr_from_webdriver_or_webelement__('mobile')

    @property
    def title(self):
        return self.__getattr_from_webdriver_or_webelement__('title')

    @property
    def name(self):
        return self.__getattr_from_webdriver_or_webelement__('name')

    @property
    def current_url(self):
        return self.__getattr_from_webdriver_or_webelement__('current_url')

    @property
    def page_source(self):
        return self.__getattr_from_webdriver_or_webelement__('page_source')

    @property
    def current_window_handle(self):
        return self.__getattr_from_webdriver_or_webelement__('current_window_handle')

    @property
    def window_handles(self):
        return self.__getattr_from_webdriver_or_webelement__('window_handles')

    @property
    def switch_to(self):
        return self.__getattr_from_webdriver_or_webelement__('switch_to')

    @property
    def desired_capabilities(self):
        return self.__getattr_from_webdriver_or_webelement__('desired_capabilities')

    @property
    def file_detector(self):
        return self.__getattr_from_webdriver_or_webelement__('file_detector')

    @file_detector.setter
    def file_detector(self, detector):
        self.__setattr_to_webdriver_or_webelement__('file_detector', detector)

    @property
    def orientation(self):
        return self.__getattr_from_webdriver_or_webelement__('orientation')

    @orientation.setter
    def orientation(self, value):
        self.__setattr_to_webdriver_or_webelement__('orientation', value)

    @property
    def application_cache(self):
        return self.__getattr_from_webdriver_or_webelement__('application_cache')

    @property
    def log_types(self):
        return self.__getattr_from_webdriver_or_webelement__('log_types')

    def start_client(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('start_client')(*args, **kwargs)

    def stop_client(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('stop_client')(*args, **kwargs)

    def start_session(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('start_session')(*args, **kwargs)

    def _wrap_value(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('_wrap_value')(*args, **kwargs)

    def create_web_element(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('create_web_element')(*args, **kwargs)

    def _unwrap_value(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('_unwrap_value')(*args, **kwargs)

    def execute(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('execute')(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get')(*args, **kwargs)

    def go_to(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get')(*args, **kwargs)

    def find_element_by_id(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_id')(*args, **kwargs)

    def find_elements_by_id(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_id')(*args, **kwargs)

    def find_element_by_xpath(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_xpath')(*args, **kwargs)

    def find_elements_by_xpath(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_xpath')(*args, **kwargs)

    def find_element_by_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_link_text')(*args, **kwargs)

    def find_elements_by_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_link_text')(*args, **kwargs)

    def find_element_by_partial_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_partial_link_text')(*args, **kwargs)

    def find_elements_by_partial_link_text(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_partial_link_text')(*args, **kwargs)

    def find_element_by_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_name')(*args, **kwargs)

    def find_elements_by_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_name')(*args, **kwargs)

    def find_element_by_tag_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_tag_name')(*args, **kwargs)

    def find_elements_by_tag_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_tag_name')(*args, **kwargs)

    def find_element_by_class_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_class_name')(*args, **kwargs)

    def find_elements_by_class_name(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_class_name')(*args, **kwargs)

    def find_element_by_css_selector(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element_by_css_selector')(*args, **kwargs)

    def find_elements_by_css_selector(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements_by_css_selector')(*args, **kwargs)

    def execute_script(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('execute_script')(*args, **kwargs)

    def execute_async_script(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('execute_async_script')(*args, **kwargs)

    def close(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('close')(*args, **kwargs)

    def quit(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('quit')(*args, **kwargs)

    def maximize_window(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('maximize_window')(*args, **kwargs)

    def switch_to_active_element(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('switch_to_active_element')(*args, **kwargs)

    def switch_to_window(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('switch_to_window')(*args, **kwargs)

    def switch_to_frame(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('switch_to_frame')(*args, **kwargs)

    def switch_to_default_content(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('switch_to_default_content')(*args, **kwargs)

    def switch_to_alert(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('switch_to_alert')(*args, **kwargs)

    def back(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('back')(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('forward')(*args, **kwargs)

    def refresh(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('refresh')(*args, **kwargs)

    def get_cookies(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_cookies')(*args, **kwargs)

    def get_cookie(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_cookie')(*args, **kwargs)

    def delete_cookie(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('delete_cookie')(*args, **kwargs)

    def delete_all_cookies(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('delete_all_cookies')(*args, **kwargs)

    def add_cookie(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('add_cookie')(*args, **kwargs)

    def implicitly_wait(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('implicitly_wait')(*args, **kwargs)

    def set_script_timeout(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('set_script_timeout')(*args, **kwargs)

    def set_page_load_timeout(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('set_page_load_timeout')(*args, **kwargs)

    def find_element(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_element')(*args, **kwargs)

    def find_elements(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('find_elements')(*args, **kwargs)

    def get_screenshot_as_file(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_screenshot_as_file')(*args, **kwargs)

    def save_screenshot(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('save_screenshot')(*args, **kwargs)

    def get_screenshot_as_png(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_screenshot_as_png')(*args, **kwargs)

    def get_screenshot_as_base64(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_screenshot_as_base64')(*args, **kwargs)

    def set_window_size(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('set_window_size')(*args, **kwargs)

    def get_window_size(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_window_size')(*args, **kwargs)

    def set_window_position(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('set_window_position')(*args, **kwargs)

    def get_window_position(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_window_position')(*args, **kwargs)

    def get_log(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('get_log')(*args, **kwargs)


class ChromeWebDriverInterface(BaseInterface):
    """
    Specific methods of chrome driver
    """

    def launch_app(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('launch_app')(*args, **kwargs)


class OperaWebDriverInterface(BaseInterface):
    """
    Specific methods of opera driver
    """
    pass


class FirefoxWebDriverInterface(BaseInterface):
    """
    Specific methods of firefox driver
    """

    @property
    def firefox_profile(self):
        return self.__getattr_from_webdriver_or_webelement__('firefox_profile')

    def set_context(self, *args, **kwargs):
        return self.__getattr_from_webdriver_or_webelement__('set_context')(*args, **kwargs)


class PhantomJSWebDriverInterface(BaseInterface):
    """
    Specific methods of phantomjs driver
    """
    pass


class IEWebDriverInterface(BaseInterface):
    """
    Specific methods of ie driver
    """
    pass


class WebDriverInterface(CommonWebDriverInterface,
                         ChromeWebDriverInterface,
                         OperaWebDriverInterface,
                         FirefoxWebDriverInterface,
                         PhantomJSWebDriverInterface,
                         IEWebDriverInterface):
    """
    Built interface for web driver.
    Class implemented provider interface
    to web driver instance for proxy object.
    """
    pass



# -*- coding: utf-8 -*-

from warnings import warn
from optparse import OptionGroup

from selenium.webdriver.common.keys import Keys as keys

from . import forms
from . import polling

from .case import assertion
from .case import require_browser
from .case import SeleniumAssertion
from .case import make_with_browsers
from .case import SeleniumCase as Case

from .suite import SeleniumSuite as Suite

from .proxy.actions import Alert
from .proxy.actions import ActionChains
from .proxy.actions import TouchActions

from .pageobject import Page
from .pageobject import PageItem
from .pageobject import PageElement

from .query import query

from .router import add_route

from .utils import re_raise_exc

from .browser import change_config as change_browser_config


CONFIG_KEY = 'SELENIUM_EX'


def __add_options__(parser):
    group = OptionGroup(parser, 'Selenium extension options')

    group.add_option(
        '--selenium-browser',
        dest='SELENIUM_BROWSERS',
        action='append',
        default=[],
        help='Browser name for run cases.',
    )
    group.add_option(
        '--selenium-project-url',
        dest='SELENIUM_PROJECT_URL',
        default=None,
        help='Base URL of your project.',
    )
    group.add_option(
        '--selenium-remote',
        dest='SELENIUM_REMOTE',
        action='store_true',
        default=False,
        help='Use remote server only.'
    )
    group.add_option(
        '--selenium-no-remote',
        dest='SELENIUM_NO_REMOTE',
        action='store_true',
        default=False,
        help='Use local drivers only.'
    )
    group.add_option(
        '--selenium-polling',
        dest='SELENIUM_POLLING',
        type=float,
        default=None,
        help='Polling timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-polling-delay',
        dest='SELENIUM_POLLING_DELAY',
        type=float,
        default=None,
        help='Polling delay. Float or integer value.',
    )
    group.add_option(
        '--selenium-wait-timeout',
        dest='SELENIUM_WAIT_TIMEOUT',
        type=float,
        default=None,
        help='Implicitly wait timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-page-load-timeout',
        dest='SELENIUM_PAGE_LOAD_TIMEOUT',
        type=float,
        default=None,
        help='Load page timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-script-timeout',
        dest='SELENIUM_SCRIPT_TIMEOUT',
        type=float,
        default=None,
        help='Execute script timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-window-size',
        dest='SELENIUM_WINDOW_SIZE',
        default=None,
        help='Window size. For example: 340x480',
    )

    parser.add_option_group(group)


def __install__(program):
    from .extension import EX_NAME
    from .extension import Selenium

    config = program.config.get(CONFIG_KEY, {})

    if program.config.SELENIUM_PROJECT_URL:
        config['PROJECT_URL'] = program.config.SELENIUM_PROJECT_URL

    if program.config.SELENIUM_NO_REMOTE:
        config['USE_REMOTE'] = False

    if program.config.SELENIUM_REMOTE:
        config['USE_REMOTE'] = True

    if program.config.SELENIUM_POLLING:
        config['POLLING_TIMEOUT'] = program.config.SELENIUM_POLLING

    if program.config.SELENIUM_POLLING_DELAY:
        config['POLLING_DELAY'] = program.config.SELENIUM_POLLING_DELAY

    if program.config.SELENIUM_WAIT_TIMEOUT:
        config['WAIT_TIMEOUT'] = program.config.SELENIUM_WAIT_TIMEOUT

    if program.config.SELENIUM_PAGE_LOAD_TIMEOUT:
        config['PAGE_LOAD_TIMEOUT'] = program.config.SELENIUM_PAGE_LOAD_TIMEOUT

    if program.config.SELENIUM_SCRIPT_TIMEOUT:
        config['SCRIPT_TIMEOUT'] = program.config.SELENIUM_SCRIPT_TIMEOUT

    if program.config.SELENIUM_WINDOW_SIZE:
        w, h = program.config.SELENIUM_WINDOW_SIZE.split('x')
        config['WINDOW_SIZE'] = (int(w), int(h))

    if (config.get('POLLING_TIMEOUT') and config.get('IMPLICITLY_WAIT')) \
            and config.get('POLLING_TIMEOUT') >= config.get('IMPLICITLY_WAIT'):
        warn_message = 'POLLING_TIMEOUT >= IMPLICITLY_WAIT it will be working so slowly. ' \
                       'Please set IMPLICITLY_WAIT larger POLLING_TIMEOUT'
        warn(warn_message, RuntimeWarning)

    program.shared_extension(
        EX_NAME, Selenium, args=(config, ),
    )


__all__ = (
    'Case',
    'Page',
    'keys',
    'forms',
    'query',
    'Alert',
    'Suite',
    'polling',
    'PageItem',
    'add_route',
    'assertion',
    'waiting_for',
    'PageElement',
    'ActionChains',
    'TouchActions',
    're_raise_exc',
    'inject_driver',
    'SeleniumAssertion',
    'make_with_browsers',
    'change_browser_config',
)



# -*- coding: utf-8 -*-

from optparse import OptionGroup

from .mock import BaseMock
from .base import BaseMockServer
from .json_api_mock import JsonMock
from .json_api_mock import JsonApiMockServer

from ...exceptions import EmergencyStop


EX_NAME = 'mocker'
CONFIG_KEY = 'MOCKER_EX'
DEFAULT_SERVER_TYPE = 'json_api'

SERVER_TYPES = {
    'simple': BaseMockServer,
    'json_api': JsonApiMockServer,
}


def create_server(host=None,
                  port=None,
                  mocks=None,
                  debug=False,
                  mocks_path=None,
                  server_type=DEFAULT_SERVER_TYPE,
                  **kwargs):
    try:
        mock_server_class = SERVER_TYPES[server_type]
    except KeyError:
        raise EmergencyStop(
            'Incorrect server type "{}"'.format(server_type),
        )

    return mock_server_class(
        mocks_path,
        host=host,
        port=port,
        mocks=mocks,
        debug=debug,
        **kwargs
    )


def __add_options__(parser):
    group = OptionGroup(parser, 'MockServer extension options')

    group.add_option(
        '--mock-server-mocks-dir',
        dest='MOCK_SERVER_MOCKS_DIR',
        default=None,
        help='Path to dir within mock files.'
    )
    group.add_option(
        '--mock-server-host',
        dest='MOCK_SERVER_HOST',
        default=None,
        help='Server host.',
    )
    group.add_option(
        '--mock-server-port',
        dest='MOCK_SERVER_PORT',
        default=None,
        type=int,
        help='Server port.',
    )
    group.add_option(
        '--mock-server-debug',
        dest='MOCK_SERVER_DEBUG',
        action='store_true',
        default=False,
        help='Use debug.',
    )

    parser.add_option_group(group)


def __install__(program):
    params = program.config.get(CONFIG_KEY, {})

    ex_kwargs = {
        'mocks': params.get('MOCKS'),
        'host': program.config.MOCK_SERVER_HOST or params.get('HOST'),
        'port': program.config.MOCK_SERVER_PORT or params.get('PORT'),
        'debug': program.config.MOCK_SERVER_DEBUG or params.get('DEBUG'),
        'mocks_path': program.config.MOCK_SERVER_MOCKS_DIR or params.get('MOCKS_PATH'),
        'server_type': params.get(
            'SERVER_TYPE', DEFAULT_SERVER_TYPE,
        ),
        'multiprocessing': program.config.MULTIPROCESSING,
        'threading': program.config.THREADING,
        'gevent': program.config.GEVENT,
    }

    program.shared_extension(
        EX_NAME, create_server, singleton=True, kwargs=ex_kwargs,
    )


__all__ = (
    'BaseMock',
    'JsonMock',
    'SERVER_TYPES',
    'BaseMockServer',
    'JsonApiMockServer',
)



# -*- coding: utf-8 -*-

from .extension import EX_NAME
from .extension import DBClient
from .exceptions import ConfigurationError


CONFIG_KEY = 'ALCHEMY_EX'


def __install__(program):
    config = program.config.get(CONFIG_KEY, {})

    if config:
        from . import install

        default_host = config.get('HOST')
        default_port = config.get('PORT')
        default_user = config.get('USER')
        default_password = config.get('PASSWORD')
        default_protocol = config.get('PROTOCOL')
        default_dns_params = config.get('DNS_PARAMS')
        default_pool_class = config.get('POOL_CLASS')

        session_config = config.get('SESSION', {})
        databases_config = config.get('DATABASES', {})

        for db_name, db_config in databases_config.items():
            protocol = db_config.pop('protocol', default_protocol)
            if not protocol:
                raise ConfigurationError(
                    '"protocol" is required param for database "{}"'.format(db_name),
                )

            host = db_config.pop('host', default_host)
            if not host:
                raise ConfigurationError(
                    '"host" is required param for database "{}"'.format(db_name),
                )

            port = db_config.pop('port', default_port)
            if not port:
                raise ConfigurationError(
                    '"port" is required param or database "{}"'.format(db_name),
                )

            user = db_config.pop('user', default_user)
            if not user:
                raise ConfigurationError(
                    '"user" is required param for database "{}"'.format(db_name),
                )

            password = db_config.pop('password', default_password)
            dns_params = db_config.pop('dns_params', default_dns_params)
            pool_class = db_config.pop('pool_class', default_pool_class)

            install.setup_engine(
                protocol,
                host,
                port,
                db_name,
                user,
                password=password,
                dns_params=dns_params,
                pool_class=pool_class,
                **db_config
            )

        install.setup_session(**session_config)

        program.shared_extension(EX_NAME, DBClient)
    else:
        def extension_mock():
            raise ConfigurationError(
                'Alchemy extension is not configured. Use config key "{}" for this.'.format(
                    CONFIG_KEY,
                ),
            )

        program.shared_extension(EX_NAME, extension_mock)


# -*- coding: utf-8 -*-

from copy import deepcopy

from .exceptions import ExtensionNotFound


_TMP = {}
_WAS_CLEAR = False


def install(ext, program):
    if getattr(ext, '__install__', None):
        ext.__install__(program)


def add_options(ext, parser):
    if getattr(ext, '__add_options__', None):
        ext.__add_options__(parser)


class ExtensionContainer(object):

    def __init__(self, ext, args=None, kwargs=None):
        self.__ext = ext
        self.__args = args or tuple()
        self.__kwargs = kwargs or dict()

    def __call__(self):
        return self.__ext(*self.__args, **self.__kwargs)

    @property
    def ext(self):
        return self.__ext

    @property
    def args(self):
        return self.__args

    @property
    def kwargs(self):
        return self.__kwargs


class SingletonExtensionContainer(ExtensionContainer):

    def __init__(self, *args, **kwargs):
        super(SingletonExtensionContainer, self).__init__(*args, **kwargs)

        self.__instance = None

    def __call__(self):
        if self.__instance is None:
            self.__instance = super(
                SingletonExtensionContainer, self,
            ).__call__()
        return self.__instance


def get(name):
    try:
        container = _TMP[name]
    except KeyError:
        if _WAS_CLEAR:
            raise RuntimeError(
                'Extension tmp was cleared',
            )
        raise ExtensionNotFound(name)

    if isinstance(container, ExtensionContainer):
        return container()

    return deepcopy(container)


def set(ext, name, is_data=False, singleton=False, args=None, kwargs=None):
    if is_data:
        _TMP[name] = ext
    else:
        if singleton:
            _TMP[name] = SingletonExtensionContainer(
                ext, args=args, kwargs=kwargs,
            )
        else:
            _TMP[name] = ExtensionContainer(
                ext, args=args, kwargs=kwargs,
            )


def clear():
    global _WAS_CLEAR

    _TMP.clear()
    _WAS_CLEAR = True



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from .. import runnable
from ..case import CaseBox
from ..xunit import XUnitData
from ..utils.common import waiting_for
from ..groups import get_pool_size_of_value


MPProcess = mp_manager = None


def import_mp():
    global MPProcess, mp_manager

    from multiprocessing import Process
    from multiprocessing import Manager

    MPProcess = Process
    mp_manager = Manager()


def target(suite, mp_result):
    suite(mp_result)
    mp_result.save_result()


class MPResult(object):

    MATCH = {}

    def __init__(self, result):
        self.result = result
        self.queue = mp_manager.Queue()

        self.result.support_mp(mp_manager)

    def __getattr__(self, item):
        return getattr(self.result, item)

    @staticmethod
    def pack_result_storage(storage):
        return [
            (runnable.id, xunit_data.to_marshal())
            for runnable, xunit_data in storage
        ]

    def unpack_result_storage(self, storage):
        for runnable_id, xunit_data in storage:
            yield self.MATCH[runnable_id], XUnitData.from_marshal(xunit_data)

    def match(self, suite):
        self.MATCH[suite.id] = suite
        suite.support_mp(mp_manager)

        for case in suite:
            if isinstance(case, CaseBox):
                for c in case:
                    self.MATCH[c.id] = c
                    c.support_mp(mp_manager)
            else:
                self.MATCH[case.id] = case
                case.support_mp(mp_manager)

    def save_result(self):
        if self.result.proxies:
            self.queue.put(
                (
                    self.result.proxies[0].name,
                    self.result.proxies[0].runtime,

                    self.pack_result_storage(
                        self.result.successes,
                    ),
                    self.pack_result_storage(
                        self.result.skipped,
                    ),
                    self.pack_result_storage(
                        self.result.failures,
                    ),
                    self.pack_result_storage(
                        self.result.errors,
                    ),
                ),
            )

    def sync(self):
        while not self.queue.empty():
            name, runtime, successes, skipped, failures, errors = self.queue.get()

            result_proxy = self.create_proxy(
                name=name,
            )
            result_proxy.runtime = runtime

            result_proxy.errors.extend(
                self.unpack_result_storage(errors),
            )
            result_proxy.successes.extend(
                self.unpack_result_storage(successes),
            )
            result_proxy.skipped.extend(
                self.unpack_result_storage(skipped),
            )
            result_proxy.failures.extend(
                self.unpack_result_storage(failures),
            )

            self.result.extend(result_proxy)
            self.result.proxies.append(result_proxy)


class Multiprocessing(object):

    def __init__(self, result, config, suites=None):
        self.queue = []
        self.stack = []

        self.mp_result = MPResult(result)
        self.release_timeout = config.MULTIPROCESSING_TIMEOUT
        self.max_processes = get_pool_size_of_value(config.ASYNC_SUITES)

        if suites:
            self.add_suites(suites)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.terminate_all()
        self.join_all()
        self.mp_result.sync()

    def add_suite(self, suite):
        self.mp_result.match(suite)

        self.queue.append(
            MPProcess(
                target=target,
                args=(suite, self.mp_result),
            ),
        )

    def add_suites(self, suites):
        for suite in suites:
            self.add_suite(suite)

    def try_release(self):
        for process in self.stack[::-1]:
            if not process.is_alive():
                process.join(timeout=self.release_timeout)
                self.stack.remove(process)

    def is_release(self):
        self.try_release()
        return len(self.stack) < self.max_processes

    def wait_release(self):
        waiting_for(
            self.is_release,
            timeout=self.release_timeout,
            message='Process list has not been release for "{}" sec.'.format(
                self.release_timeout,
            ),
        )

    def join_all(self):
        for process in self.stack:
            process.join(timeout=self.release_timeout)

    def terminate_all(self):
        for process in self.stack:
            process.terminate()

    def serve(self):
        while self.queue:
            process = self.queue.pop()
            process.start()
            self.stack.append(process)
            self.wait_release()

        self.join_all()


class MultiprocessingSuiteGroup(runnable.RunnableGroup):

    def __run__(self, result):
        self._is_run = True

        import_mp()

        with Multiprocessing(result, self.config, suites=self.objects) as mp:
            mp.serve()



# -*- coding: utf8 -*-

import time
import socket
import logging
from functools import wraps

try:
    from httplib import HTTPException
except ImportError:  # please python 3
    from http.client import HTTPException

from selenium.common.exceptions import WebDriverException

from ...utils import pyv
from .exceptions import PollingTimeoutExceeded


logger = logging.getLogger(__name__)


POLLING_EXCEPTIONS = (
    IOError,
    OSError,
    HTTPException,
    WebDriverException,
)
DEFAULT_POLLING_TIMEOUT = 30
DEFAULT_POLLING_DELAY = None


def do(callback,
       exceptions=POLLING_EXCEPTIONS,
       timeout=DEFAULT_POLLING_TIMEOUT,
       delay=DEFAULT_POLLING_DELAY):
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            exc = None
            t_start = time.time()

            while time.time() <= t_start + timeout:
                try:
                    logger.debug(
                        u'Do polling, try to call "{}", args={}, kwargs={}'.format(
                            pyv.get_func_name(f), pyv.unicode(args), pyv.unicode(kwargs),
                        ),
                    )
                    return f(*args, **kwargs)
                except socket.error:  # if connection will be refused,
                    # then we not need to continue polling
                    raise
                except exceptions as error:
                    exc = error
                    if delay:
                        time.sleep(delay)
                    continue
            else:
                if exc:
                    raise
                raise PollingTimeoutExceeded(str(timeout))

        return wrapped
    return wrapper(callback)


def wrap(timeout=DEFAULT_POLLING_TIMEOUT,
         delay=DEFAULT_POLLING_DELAY,
         exceptions=POLLING_EXCEPTIONS):
    def wrapper(f):
        return do(f,
                  timeout=timeout,
                  delay=delay,
                  exceptions=exceptions)
    return wrapper



# -*- coding: utf-8 -*-

from functools import wraps

from selenium.common.exceptions import NoSuchElementException

from ....utils import pyv
from ..query import make_result
from ..exceptions import FieldError
from ..utils import declare_standard_callback


def _if_value_is_int_then_to_string(value):
    if type(value) is int:
        return str(value)
    return value


def selector(**kwargs):
    """
    proxy for tag attributes
    """
    return kwargs


def fill_field_handler(f):
    @wraps(f)
    def wrapper(self, value=None):
        if self.value is None and value is None:
            return

        if callable(self.before_fill_trigger):
            self.before_fill_trigger(self)

        result = f(self, value=value)

        if callable(self.after_fill_trigger):
            self.after_fill_trigger(self)

        self.group.fill_memo.add(self)

        return result
    return wrapper


def clear_field_handler(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        result = f(self, *args, **kwargs)

        try:
            self.group.fill_memo.remove(self)
        except KeyError:
            pass

        return result
    return wrapper


class SimpleFieldInterface(object):

    @property
    def weight(self):
        raise NotImplementedError(
            'Property "weight" not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )

    def fill(self, value=None):
        raise NotImplementedError(
            'Method "fill" not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )

    def clear(self):
        raise NotImplementedError(
            'Method "clear" not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )


class FormField(object):
    """
    Base class for all fields
    """

    __tag__ = None

    def __init__(self,
                 name,
                 value=None,
                 group=None,
                 weight=None,
                 selector=None,
                 required=False,
                 error_mess=None,
                 invalid_value=None,
                 before_fill_trigger=None,
                 after_fill_trigger=None,
                 area=None):

        self.name = name

        self.value = value
        self.required = required
        self.error_mess = error_mess
        self.invalid_value = invalid_value

        self.__area = area
        self.__group = group

        self.__weight = weight
        self.__selector = selector or {}

        self.before_fill_trigger = declare_standard_callback(before_fill_trigger)
        self.after_fill_trigger = declare_standard_callback(after_fill_trigger)

    def __getattr__(self, item):
        return getattr(self.we, item)

    def __call__(self, group, **kwargs):
        kwargs.setdefault('area', self.__area)
        kwargs.setdefault('weight', self.__weight)
        kwargs.setdefault('selector', self.selector)
        kwargs.setdefault('required', self.required)
        kwargs.setdefault('error_mess', self.error_mess)
        kwargs.setdefault('before_fill_trigger', self.before_fill_trigger)
        kwargs.setdefault('after_fill_trigger', self.after_fill_trigger)
        kwargs.setdefault('value', self.value() if callable(self.value) else self.value)
        kwargs.setdefault(
            'invalid_value', self.invalid_value() if callable(self.invalid_value) else self.invalid_value,
        )

        return self.__class__(
            self.name,
            group=group,
            **kwargs
        )

    @property
    def browser(self):
        return self.group.browser

    @property
    def area(self):
        if self.__area:
            return self.__area(self.group.area)
        return self.group.area

    @property
    def group(self):
        if not self.__group:
            raise RuntimeError(
                'Field is not binding to group',
            )
        return self.__group

    @property
    def weight(self):
        return self.__weight

    @property
    def selector(self):
        return self.__selector

    @property
    def we(self):
        return make_result(
            self.area,
            self.__tag__,
        )(**self.__selector).first()

    @property
    def attr(self):
        return self.we.attr

    @property
    def css(self):
        return self.we.css


class Input(FormField, SimpleFieldInterface):

    __tag__ = 'input'

    @fill_field_handler
    def fill(self, value=None):
        value = _if_value_is_int_then_to_string(value or self.value)

        self.we.send_keys(*value)

    @clear_field_handler
    def clear(self):
        self.we.clear()


assert issubclass(Input, SimpleFieldInterface)


class TextArea(Input):

    __tag__ = 'textarea'


assert issubclass(TextArea, SimpleFieldInterface)


class DateInput(Input):

    __default_format__ = '%d.%m.%Y'

    def __init__(self, name, **kwargs):
        self.format = kwargs.pop('format', self.__default_format__)

        super(DateInput, self).__init__(name, **kwargs)

    def __call__(self, group, **kwargs):
        kwargs.setdefault('format', self.format)
        return super(DateInput, self).__call__(group, **kwargs)

    def _get_value(self, value):
        date = value or self.value
        if isinstance(date, pyv.basestring):
            return date
        return date.strftime(self.format)

    def fill(self, value=None):
        return super(DateInput, self).fill(
            self._get_value(value),
        )


assert issubclass(DateInput, Input)


class Checkbox(FormField, SimpleFieldInterface):

    __tag__ = 'input'

    def __init__(self,
                 name,
                 **kwargs):
        self.force = kwargs.pop('force', False)

        super(Checkbox, self).__init__(name, **kwargs)

    def __call__(self, group, **kwargs):
        kwargs.setdefault('force', self.force)
        return super(Checkbox, self).__call__(group, **kwargs)

    @fill_field_handler
    def fill(self, value=None):
        value = value or self.value

        el = self.we
        current_value = el.is_selected()

        if self.group.__allow_raises__:

            if current_value and value:
                raise FieldError('Oops, checkbox was selected')
            elif not current_value and not value:
                raise FieldError('Oops, checkbox was unselected')

        if (value and not current_value) or (not value and current_value):
            if self.force:
                while not el.is_selected():
                    el.click()
            else:
                el.click()

    @clear_field_handler
    def clear(self):
        el = self.we

        if el.is_selected():
            el.click()


assert issubclass(Checkbox, SimpleFieldInterface)


class RadioButton(Checkbox):

    __tag__ = 'input'

    @fill_field_handler
    def fill(self, value=None):
        value = value or self.value

        if value is None:
            return False

        el = self.we
        current_value = el.is_selected()
        changed = False

        if (value and not current_value) or (not value and current_value):
            if self.force:
                while not el.is_selected():
                    el.click()
            else:
                el.click()
            changed = True

        return changed

    @clear_field_handler
    def clear(self):
        pass


assert issubclass(RadioButton, SimpleFieldInterface)


class Select(FormField, SimpleFieldInterface):

    __tag__ = 'select'

    FILL_BY_TEXT = 'by text'
    FILL_BY_VALUE = 'by value'

    def __init__(self,
                 name,
                 **kwargs):
        self.fill_strategy = kwargs.pop('fill_strategy', self.FILL_BY_VALUE)

        super(Select, self).__init__(name, **kwargs)

    def __call__(self, group, **kwargs):
        kwargs.setdefault('fill_strategy', self.fill_strategy)
        return super(Select, self).__call__(group, **kwargs)

    @fill_field_handler
    def fill(self, value=None):
        value = _if_value_is_int_then_to_string(value or self.value)

        filters = {}
        if self.fill_strategy == self.FILL_BY_TEXT:
            filters['text'] = value
        elif self.fill_strategy == self.FILL_BY_VALUE:
            filters['value'] = value
        else:
            raise ValueError(
                'Incorrect value from "fill_strategy": "{}"'.format(self.fill_strategy),
            )

        option = self.we.option().all().get_by(**filters)

        if option:
            option.click()
        else:
            raise NoSuchElementException(
                u'Can not select value "{}" of field "{}", selector "{}"'.format(
                    value, self.name, str(self.selector),
                ),
            )

    @clear_field_handler
    def clear(self):
        pass


assert issubclass(Select, SimpleFieldInterface)



# -*- coding: utf-8 -*-

from collections import Iterator


class FieldsIterator(Iterator):

    def __init__(self, group, exclude=None):
        exclude = exclude or tuple()

        self.__current_index = 0
        self.__fields = [
            field for field in group.fields
            if field not in exclude
        ]

    def next(self):
        try:
            field = self.__fields[self.__current_index]
        except IndexError:
            self.__current_index = 0
            raise StopIteration

        self.__current_index += 1

        return field


class FieldsWithContainsInvalidValueIterator(Iterator):

    def __init__(self, group, exclude=None):
        exclude = exclude or tuple()

        self.__current_index = 0
        self.__fields = [
            field for field in group.fields
            if field not in exclude and field.invalid_value is not None
        ]

    def next(self):
        try:
            field = self.__fields[self.__current_index]
        except IndexError:
            self.__current_index = 0
            raise StopIteration

        self.__current_index += 1

        return field


class RequiredFieldsIterator(Iterator):

    def __init__(self, group, exclude=None):
        exclude = exclude or tuple()

        self.__current_index = 0
        self.__fields = [
            field for field in group.fields
            if field.required and field not in exclude
        ]

    def next(self):
        try:
            field = self.__fields[self.__current_index]
        except IndexError:
            self.__current_index = 0
            raise StopIteration

        self.__current_index += 1

        return field



# -*- coding: utf-8 -*-

from ...query import Query


ANY_TAG_METHOD_NAME = 'any'


# Tag's names when is allowed as attribute name on proxy object instance
HTML_TAGS_ALLOWED_AS_METHOD = (
    Query.A, Query.B, Query.P, Query.U, Query.UL,
    Query.LI, Query.BR, Query.EM, Query.HR, Query.TR,
    Query.TD, Query.TH, Query.TT, Query.VAR, Query.IMG,
    Query.DIV, Query.MAP, Query.HEAD, Query.FORM, Query.BODY,
    Query.AREA, Query.CODE, Query.BASE, Query.SPAN, Query.LINK,
    Query.META, Query.SMALL, Query.TABLE, Query.INPUT, Query.LABEL,
    Query.FRAME, Query.EMBED, Query.BLINK, Query.IFRAME, Query.CENTER,
    Query.STRONG, Query.BUTTON, Query.OBJECT, Query.OPTION, Query.SELECT,
    Query.TEXTAREA, Query.OPTGROUP, Query.FRAMESET, ANY_TAG_METHOD_NAME,
)


class BaseInterface(object):
    """
    Base class to interface for proxy object
    """

    def __getattr_from_webdriver_or_webelement__(self, item):
        raise NotImplementedError(
            'Method "__getattr_from_webdriver_or_webelement__" does not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )

    def __setattr_to_webdriver_or_webelement__(self, item, value):
        raise NotImplementedError(
            'Method "__setattr_to_webdriver_or_webelement__" does not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )



# -*- coding: utf-8 -*-

from ..utils import change_name_from_python_to_html


class WebElementToObject(object):
    """
    Get a chance for work with attributes
    of HTML tag as attributes of the object
    """

    def __init__(self, proxy, allow_raise=True):
        self.__dict__['__proxy__'] = proxy
        self.__dict__['__allow_raise__'] = allow_raise

    def __getattr__(self, item):
        atr = self.__dict__['__proxy__'].get_attribute(
            change_name_from_python_to_html(item),
        )

        if atr:
            return atr

        if self.__dict__['__allow_raise__']:
            raise AttributeError(item)

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].setAttribute(arguments[1], arguments[2]);',
            self.__dict__['__proxy__']._wrapped,
            change_name_from_python_to_html(key),
            value,
        )


class WebElementCssToObject(object):
    """
    Like WebElementToObject but work with css attributes
    """

    def __init__(self, proxy):
        self.__dict__['__proxy__'] = proxy

    def __getattr__(self, item):
        return self.__dict__['__proxy__'].value_of_css_property(
            change_name_from_python_to_html(item),
        )

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].style[arguments[1]] = arguments[2];',
            self.__dict__['__proxy__']._wrapped,
            change_name_from_python_to_html(key),
            value,
        )



# -*- coding: utf-8 -*-

from . import collector
from .rule import SettingsValue
from .staging import AliasToMethod
from .collector import current_schema
from .signature import Signature as sig
from .exceptions import ConfigurationError


class BuilderSchema(object):

    def __init__(self, name, originated=None, factory=None):
        self.__name = name
        self.__factory = factory
        self.__originated = originated

    def __getattr__(self, item):
        try:
            return getattr(self.__factory, item)
        except AttributeError:
            pass

        try:
            return getattr(self.__originated, item)
        except AttributeError:
            pass

        raise AttributeError('{}({}): {}'.format(self.__class__.__name__, self.__name, item))

    def __contains__(self, item):
        return item in self.__dict__

    def __getitem__(self, item):
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __repr__(self):
        if self.__originated and self.__factory:
            return '<Schema({} spawned {}): {}>'.format(
                self.__factory.__class__.__name__,
                self.__originated.__class__.__name__, self.__name,
            )
        return '<Schema: {}>'.format(self.__name)

    def __dir__(self):
        self_dir = [k for k in self.__dict__ if not k.startswith('_')]

        return list(
            set(
                dir(self.__originated) + dir(self.__factory) + self_dir
            ),
        )


class BaseBuilderSettings(object):

    def __init__(self, **options):
        self.__options = options

    @property
    def options(self):
        return self.__options


class BaseBuilder(object):

    __build_schema__ = {}
    __schema_class__ = BuilderSchema
    __settings_class__ = BaseBuilderSettings

    def __init__(self, runnable=None):
        self._schema = None
        self._settings = None
        self._runnable = runnable

    def __repr__(self):
        return '<Builder on: {}>'.format(
            repr(self._runnable).strip('<>'),
        )

    @property
    def schema(self):
        return self._schema

    @property
    def settings(self):
        return self._settings

    @property
    def runnable(self):
        return self._runnable

    def configure(self, **kwargs):
        self._settings = self.__settings_class__(**kwargs)

    def build(self):
        if not self._settings:
            raise ConfigurationError(
                'Builder was not configured. Should to configure builder for this operation.',
            )

        self._schema = self.__schema_class__('general')

        collector.collect(self)


__all__ = (
    'sig',
    'BaseBuilder',
    'AliasToMethod',
    'BuilderSchema',
    'SettingsValue',
    'current_schema',
    'BaseBuilderSettings',
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from threading import Thread


class RunServerInThread(Thread):

    def __init__(self, mock_server, server_class):
        self._httpd = None
        self._mock_server = mock_server
        self._server_class = server_class

        super(RunServerInThread, self).__init__()

    def run(self):
        self._httpd = self._server_class(
            self._mock_server.host,
            self._mock_server.port,
            self._mock_server,
        )
        self._httpd.serve_forever()

    def stop(self):
        if self._httpd:
            self._httpd.shutdown()



# -*- coding: utf-8 -*-

from .webdriver import WebDriverInterface
from .webelement import WebElementInterface
from .base import HTML_TAGS_ALLOWED_AS_METHOD


__all__ = (
    'WebDriverInterface',
    'WebElementInterface',
    'HTML_TAGS_ALLOWED_AS_METHOD',
)



# -*- coding: utf-8 -*-

"""
Environments:

SIMPLE_SEISMOGRAPH - to install library without requirements of extensions
SEISMOGRAPH_EXTENSIONS - names of extensions separated by commas
"""

import os
import re

from setuptools import setup
from setuptools import find_packages


with open('seismograph/__init__.py') as _:
    __version__ = re.search(r"__version__\s*=\s*'(.*)'", _.read(), re.M).group(1)
assert __version__


EXTENSIONS = [
    'mocker',
    'alchemy',
    'builder',
    'selenium',
]

REQUIREMENTS = [
    'six>=1.4',
]

CONSOLE_SCRIPTS = [
    'seismograph = seismograph.__main__:main',
]

EX_REQUIREMENTS = {
    'mocker': [
        'requests',
        'flask>=0.7',
    ],
    'selenium': [
        'selenium>=2.46',
    ],
    'alchemy': [
        'sqlalchemy>=0.8',
    ],
    'builder': [
        'jsonschema>=1.0',
    ],
}

EX_CONSOLE_SCRIPTS = {
    'mocker': [
        'seismograph.mocker = seismograph.ext.mocker.__main__:main',
    ],
}

SIMPLE_INSTALL = os.getenv('SIMPLE_SEISMOGRAPH')
EXTENSIONS_TO_INSTALL = os.getenv('SEISMOGRAPH_EXTENSIONS')


def prepare_extension_data(ex_name):
    if ex_name not in EXTENSIONS:
        raise RuntimeError(
            'Unknown extension name: {}'.format(ex_name),
        )

    requirements = EX_REQUIREMENTS.get(ex_name)
    console_scripts = EX_CONSOLE_SCRIPTS.get(ex_name)

    if requirements:
        REQUIREMENTS.extend(requirements)

    if console_scripts:
        CONSOLE_SCRIPTS.extend(console_scripts)


def prepare_data():
    if SIMPLE_INSTALL:
        return

    if EXTENSIONS_TO_INSTALL:
        extensions_to_install = [
            n.strip()
            for n in EXTENSIONS_TO_INSTALL.split(',')
            if n.strip()
        ]
        for ex_name in extensions_to_install:
            prepare_extension_data(ex_name)
    else:
        for ex_name in EXTENSIONS:
            prepare_extension_data(ex_name)


def install_package():
    setup(
        name='seismograph',
        version=__version__,
        url='https://github.com/trifonovmixail/seismograph',
        packages=find_packages(exclude=('example*', 'tests*')),
        author='Mikhail Trifonov',
        author_email='trifonovmixail@ya.ru',
        license='GNU LGPL',
        description='Framework for test development',
        keywords='test unittest framework',
        long_description=open('README.rst').read(),
        include_package_data=True,
        zip_safe=False,
        platforms='any',
        install_requires=REQUIREMENTS,
        entry_points={
            'console_scripts': CONSOLE_SCRIPTS,
        },
        test_suite='tests',
        classifiers=(
            'Development Status :: 4 - Beta',
            'Natural Language :: Russian',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: Implementation :: CPython',
            'Topic :: Software Development :: Testing',
        ),
    )


if __name__ == '__main__':
    prepare_data()
    install_package()



# -*- coding: utf-8 -*-

import abc
from functools import wraps

from six import with_metaclass

from ...utils import pyv
from .exceptions import RestrictionError


FACTORY_METHOD_PREFIX = 'create_'


def _set_storage_name(method, storage_name):
    setattr(method, '__storage_name__', storage_name)


def _get_storage_name(method, default):
    return getattr(method, '__storage_name__', default)


def _get_default_storage_name(method_name):
    return method_name.replace(FACTORY_METHOD_PREFIX, '')


def _set_require(f, *what):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        for it in what:
            if not getattr(self, it):
                raise RestrictionError(
                    'Method "{}" from class "{}" can not be called. "{}" is required.'.format(
                        pyv.get_func_name(f),
                        self.__class__.__name__,
                        it,
                    ),
                )

        return f(self, *args, **kwargs)
    return wrapper


def _set_only_one_creation(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        method_name = pyv.get_func_name(f)
        storage_name = _get_storage_name(
            f, _get_default_storage_name(method_name),
        )

        if getattr(self, storage_name) is not None:
            raise RestrictionError(
                '"{}" already created on {}'.format(
                    storage_name, self,
                ),
            )

        return f(self, *args, **kwargs)
    return wrapper


def factory_method(**options):
    def wrapper(f):
        require = options.get('require')
        storage = options.get('storage')
        only_one_creation = options.get('only_one_creation')

        if storage:
            _set_storage_name(f, storage)

        if only_one_creation:
            f = _set_only_one_creation(f)

        if require:
            if not isinstance(require, (list, tuple)):
                require = (require, )

            f = _set_require(f, *require)

        @wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped
    return wrapper


class FactoryMeta(abc.ABCMeta):

    def __new__(mcs, *args, **kwargs):
        cls = abc.ABCMeta.__new__(mcs, *args, **kwargs)
        attributes = list(n for n in dir(cls) if not n.startswith('_'))
        factory_methods = filter(lambda n: n.startswith(FACTORY_METHOD_PREFIX), attributes)

        for method_name in factory_methods:
            storage_name = _get_storage_name(
                getattr(cls, method_name),
                _get_default_storage_name(method_name),
            )

            if storage_name not in attributes:
                raise RestrictionError(
                    'Class "{}" has method "{}" but does not has property "{}"'.format(
                        cls.__name__,
                        method_name,
                        storage_name,
                    ),
                )

        return cls


class BaseFactory(with_metaclass(FactoryMeta, object)):

    @abc.abstractproperty
    def data(self):
        pass

    @abc.abstractmethod
    def build_data(self, **data):
        pass

    @abc.abstractmethod
    def create(self, *args, **kwargs):
        pass

    @property
    def factory_methods(self):
        return sorted(
            list(
                filter(lambda n: n.startswith(FACTORY_METHOD_PREFIX), dir(self)),
            ),
        )



# -*- coding: utf-8 -*-

from ...exceptions import SeismographError


class BaseBuilderError(SeismographError):
    pass


class ConfigurationError(BaseBuilderError):
    pass


class SchemaError(BaseBuilderError):
    pass


class FactoryError(BaseBuilderError):
    pass


class RestrictionError(FactoryError):
    pass


class ValidationError(BaseBuilderError):
    pass


class BuildError(BaseBuilderError):
    pass


class RelationError(SchemaError):
    pass



# -*- coding: utf-8 -*-

from flask.app import _endpoint_from_view_func


def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
    if endpoint is None:
        endpoint = _endpoint_from_view_func(view_func)

    options['endpoint'] = endpoint
    methods = options.pop('methods', None)

    if methods is None:
        methods = getattr(view_func, 'methods', None) or ('GET',)

    methods = set(methods)
    required_methods = set(getattr(view_func, 'required_methods', ()))

    provide_automatic_options = getattr(
        view_func, 'provide_automatic_options', None,
    )

    if provide_automatic_options is None:
        if 'OPTIONS' not in methods:
            provide_automatic_options = True
            required_methods.add('OPTIONS')
        else:
            provide_automatic_options = False

    methods |= required_methods

    options['defaults'] = options.get('defaults') or None

    rule = self.url_rule_class(rule, methods=methods, **options)
    rule.provide_automatic_options = provide_automatic_options

    self.url_map.add(rule)

    if view_func is not None:
        old_func = self.view_functions.get(endpoint)

        if old_func is not None and old_func != view_func:
            raise AssertionError(
                'View function mapping is overwriting '
                'an existing endpoint function: %s' % endpoint,
            )

        self.view_functions[endpoint] = view_func
